module GameLogic exposing (..)

import List.Extra
import List exposing (map)
import Grid exposing (Grid)
import Types exposing (..)
import TypeHelperFunctions exposing (..)
import GeneralHelperFunctions exposing (..)

-- (boardx, boardy) -> (Board origin x, Board origin y)
getBoardBeginCoordinates : (Int, Int) -> (Int, Int)
getBoardBeginCoordinates (x,y) = ((1000-(x*50))//2, (500-(y*50))//2)

setCurrentLevel : Model -> Int -> Model
setCurrentLevel model levelNum =
    let
        model0 = {model | levelList = saveCurrentLevel model}
        indexres = indexLL model0.levelList levelNum
    in {model0 | saved = Tuple.first indexres,
        currentLevel = levelNum + 1,
        boardGrid = Tuple.second indexres,
        boardIsFull = False}

saveCurrentLevel : Model -> LevelList
saveCurrentLevel model = if model.currentLevel == 0  || List.length model.saved == 0 then model.levelList
    else Levels (List.Extra.setAt (model.currentLevel-1) (model.saved, model.boardGrid) (extractLevelList model.levelList))

getCoordinates : ShapeStatus -> List (Int, Int)
getCoordinates shape = case shape.sType of
    O -> [(0,0),(0,1),(1,0),(1,1)]
    I -> case abs (modBy 4 shape.v) of
        0 -> [(0,0),(1,0),(2,0),(3,0)]
        1 -> [(-1,0),(-1,1),(-1,2),(-1,3)]
        2 -> [(-1,-1),(-2,-1),(-3,-1),(-4,-1)]
        _ -> [(0,-1),(0,-2),(0,-3),(0,-4)]
    T -> case abs (modBy 4 shape.v) of
        0 -> [(0,0),(1,0),(1,-1),(2,0)]
        1 -> [(-1,0),(-1,1),(0,1),(-1,2)]
        2 -> [(-1,-1),(-2,-1),(-2,0),(-3,-1)]
        _ -> [(0,-1),(0,-2),(-1,-2),(0,-3)]
    J -> case abs (modBy 4 shape.v) of
        0 -> [(-1,0),(-2,0),(-3,0),(-3,-1)]
        1 -> [(-1,-1),(-1,-2),(-1,-3),(0,-3)]
        2 -> [(0,-1),(1,-1),(2,-1),(2,0)]
        _ -> [(0,0),(0,1),(0,2),(-1,2)]
    L -> case abs (modBy 4 shape.v) of
        0 -> [(0,0),(1,0),(2,0),(2,-1)]
        1 -> [(-1,0),(-1,1),(-1,2),(0,2)]
        2 -> [(-1,-1),(-2,-1),(-3,-1),(-3,0)]
        _ -> [(0,-1),(0,-2),(0,-3),(-1,-3)]
    Z -> case abs (modBy 4 shape.v) of
        0 -> [(-1,0),(-2,0),(-2,-1),(-3,-1)]
        1 -> [(-1,-1),(-1,-2),(0,-2),(0,-3)]
        2 -> [(0,-1),(1,-1),(1,0),(2,0)]
        _ -> [(0,0),(0,1),(-1,1),(-1,2)]
    S -> case abs (modBy 4 shape.v) of
        0 -> [(0,0),(1,0),(1,-1),(2,-1)]
        1 -> [(-1,0),(-1,1),(0,1),(0,2)]
        2 -> [(-1,-1),(-2,-1),(-2,0),(-3,0)]
        _ -> [(0,-1),(0,-2),(-1,-2),(-1,-3)]
    Custom c -> if c.canRotate then
        let
            centercoords = map adjustToCenter c.coordinates
        in case abs (modBy 4 shape.v) of
            0 -> c.coordinates
            1 -> map adjustToOrigin (map rotateTupleCw centercoords)
            2 -> map adjustToOrigin (map (rotateTupleCw >> rotateTupleCw) centercoords)
            _ -> map adjustToOrigin (map (rotateTupleCw >> rotateTupleCw >> rotateTupleCw) centercoords)
        else c.coordinates
    _ -> []

canPlace : Model -> ShapeStatus -> Int -> Int -> Bool
canPlace model shape x y =
    let
        grid = model.boardGrid
        relativeCoords = List.map (\(tx,ty) -> (x+tx,y+ty)) (getCoordinates shape)
    in List.foldr (&&) (True) (map (isValidPosition grid ) relativeCoords)

isValidPosition : Grid Bool -> (Int, Int) -> Bool
isValidPosition grid pos =
    let
        posValue = Grid.get pos grid
    in case posValue of
        Nothing -> False
        Just bVal -> not bVal
    
getBoardCoordinates : Int -> Int -> ShapeStatus -> Maybe (Int, Int)
getBoardCoordinates bw bh shape =
    let
        (tx,ty) = getXY shape
        (bx,by) = getBoardBeginCoordinates (bw,bh)
    in if tx>=(bx-20) && tx<=(bx+(50*bw)+20) && ty>=(by-20) && ty<=(by+(50*bh)+20) then
        let
            ax = tx-(bx-20)
            ay = ty-(by-20)
        in Just (ax // 50, ay // 50)
    else Nothing

getBoardCoordinatesM : Int -> Int -> (Int,Int) -> Maybe (Int, Int)
getBoardCoordinatesM bw bh (mx,my) =
    let
        (bx,by) = getBoardBeginCoordinates (bw,bh)
    in if mx>=(bx-25) && mx<=(bx+(50*bw)+25) && my>=(by-25) && my<=(by+(50*bh)+25) then
        let
            ax = mx-(bx-25)
            ay = my-(by-25)
        in Just (ax // 50, ay // 50)
    else Nothing

updateCoordinates : Int -> Int -> ShapeStatus -> Int -> Int -> ShapeStatus
updateCoordinates bw bh shape x1 y1 = 
    let
        (bx,by) = getBoardBeginCoordinates (bw,bh)
        shapeX = {shape | x = (x1*50+bx)}
    in {shapeX | y = (y1*50+by)}

placeAt : Model -> Int -> Int -> Model
placeAt model x y =
    let
        newShape = placeOnBoard (updateCoordinates (Grid.width model.boardGrid) (Grid.height model.boardGrid) model.shapeStatus x y)
        relativeCoords = List.map (\(tx,ty) -> (x+tx,y+ty)) (getCoordinates newShape)
        newGrid = wrapSet True model.boardGrid relativeCoords
        model1 = {model | shapeStatus = createEmptyShape}
        model2 = {model1 | saved=model1.saved++[newShape]}
        model3 = {model2 | boardIsFull = checkBoard newGrid}
    in {model3 | boardGrid = newGrid}

wrapSet : Bool -> Grid Bool -> List (Int, Int) -> Grid Bool
wrapSet val grid tuples = case List.head tuples of
    Nothing -> grid
    Just tuple -> case List.tail tuples of
        Just tupleTail -> wrapSet val (Grid.set tuple val grid) tupleTail
        Nothing -> grid

rotateShapeBy : ShapeStatus -> Int -> ShapeStatus
rotateShapeBy s r = {s | v = modBy 4 (s.v+r)}

updatePosition : ShapeStatus -> (Int, Int) -> ShapeStatus
updatePosition shape (x1,y1) = {shape | x = x1, y = y1}

checkBoard : Grid Bool -> Bool
checkBoard grid = Grid.foldl (&&) (True) grid

levelIncomplete : Model -> Int -> Bool
levelIncomplete model rel = if List.member (model.currentLevel+rel) model.completedLevels then False else True

rotateClockwiseBy : Model -> Int -> Model
rotateClockwiseBy model v =
    let
        x = Tuple.first model.mousePiecePosition
        y = Tuple.second model.mousePiecePosition
        model1 = {model | mousePiecePosition = (y*(-v),x*(v))}
    in update (Position (Tuple.first model1.mpos) (Tuple.second model1.mpos)) model1



update : Msg -> Model -> Model
update msg model = case msg of
    Position x y ->
        let
            model1 = {model | shapeStatus = updatePosition model.shapeStatus (addTuple (x,y) model.mousePiecePosition)}
        in {model1 | mpos = (x,y)}
    RightRotate ->
        let
            model1 = {model | shapeStatus = rotateShapeBy model.shapeStatus 1}
        in case model1.shapeStatus.sType of
            O -> model1
            Custom c -> if c.canRotate then rotateClockwiseBy model1 1 else model1
            _ -> rotateClockwiseBy model1 1
    PlaceShape -> case model.shapeStatus.sType of
        None -> model
        shape -> case getBoardCoordinates (Grid.width model.boardGrid) (Grid.height model.boardGrid) model.shapeStatus of
            Nothing -> let mmodel = {model | saved = model.saved++[model.shapeStatus]} in {mmodel | shapeStatus = createEmptyShape}
            Just (x,y) -> if canPlace model model.shapeStatus x y then placeAt model x y else model
    TakeShape i -> let mlist = List.filter (\s -> s.sIdent == i) (model.saved) in case List.head mlist of
        Just shape -> case model.shapeStatus.sType of
            None -> if shape.placed then
                    let
                        (x,y) = toJust (getBoardCoordinates (Grid.width model.boardGrid) (Grid.height model.boardGrid) shape)
                        relativeCoords = List.map (\(tx,ty) -> (x+tx,y+ty)) (getCoordinates shape)
                        newGrid = wrapSet False model.boardGrid relativeCoords
                        newShape = removeFromBoard shape
                        model1 = {model | shapeStatus = newShape}
                        model2 = {model1 | saved = (List.Extra.remove shape model.saved)}
                        model3 = {model2 | boardIsFull = checkBoard newGrid}
                        model4 = {model3 | mousePiecePosition = subTuple (getXY shape) model.mpos}
                    in {model4 | boardGrid = newGrid}

                else let
                        model1 = {model | shapeStatus = shape}
                        model2 = {model1 | mousePiecePosition = subTuple (getXY shape) model.mpos}
                    in {model2 | saved = (List.Extra.remove shape model.saved)}
            _ -> model
        Nothing -> model
    CompleteLevel -> {model | completedLevels = model.completedLevels++[model.currentLevel]}
    SwitchJsonHelp -> let state = boolToJust model.jsonHelpState in {model | jsonHelpState = Just (not state)}
    ExampleChanged s -> {model | exampleLevels = s}
    ToggleTheme -> {model | darkMode = not model.darkMode}
    DisableModal -> {model | exitModalState = False}
    EnableModal -> {model | exitModalState = True}
    ToMainMenu ->
        let
            model1 = {model | exitModalState = False}
            model2 = {model1 | jsonHelpState = if model1.jsonHelpState == Just True then Just True else Nothing}
            model3 = {model2 | currentLevel = 0}
            model4 = {model3 | completedLevels = [], drawState = Nothing}
        in {model4 | levelList = Empty}
    CheckRotate checked -> case model.levelList of
        Editor (x, ShapeCreator (cshape, index)) -> let newCshape = {cshape | canRotate = checked}
            in {model | levelList = Editor (x, ShapeCreator (newCshape, index))}
        _ -> model
    NewColour c -> case model.levelList of
        Editor (x, ShapeCreator (cshape, index)) -> let newCshape = {cshape | colour = c}
            in {model | levelList = Editor (x, ShapeCreator (newCshape, index))}
        _ -> model
    ChangeName name -> case model.levelList of
        Editor (x, ShapeCreator (cshape, index)) -> let newCshape = {cshape | shortHand = name}
            in {model | levelList = Editor (x, ShapeCreator (newCshape, index))}
        _ -> model
    DrawCoordinates -> {model | drawState = Just DCoordinates}
    DrawPolygon -> {model | drawState = Just DPolygon}
    ResetPolygon -> case model.levelList of
        Editor (x, ShapeCreator (cshape, index)) -> let newCshape = {cshape | polygonPoints = []}
            in {model | levelList = Editor (x, ShapeCreator (newCshape, index))}
        _ -> model
    AddCoordinate (cx,cy) -> case model.levelList of
        Editor (x, ShapeCreator (cshape, index)) -> let newCshape = {cshape | coordinates = cshape.coordinates++[(cx,cy)]}
            in {model | levelList = Editor (x, ShapeCreator (newCshape, index))}
        _ -> model
    RemoveCoordinate (cx,cy) -> case model.levelList of
        Editor (x, ShapeCreator (cshape, index)) -> let newCshape = {cshape | coordinates = List.Extra.remove (cx,cy) cshape.coordinates}
            in {model | levelList = Editor (x, ShapeCreator (newCshape, index))}
        _ -> model
    AddPolygonPoint (cx,cy) -> case model.levelList of
        Editor (x, ShapeCreator (cshape, index)) -> let newCshape = {cshape | polygonPoints = cshape.polygonPoints++[(cx,cy)]}
            in {model | levelList = Editor (x, ShapeCreator (newCshape, index))}
        _ -> model
    UndoPolygonPoint -> case model.levelList of
        Editor (x, ShapeCreator (cshape, index)) -> let newCshape = {cshape | polygonPoints = List.Extra.removeAt (List.length cshape.polygonPoints-1)cshape.polygonPoints}
            in {model | levelList = Editor (x, ShapeCreator (newCshape, index))}
        _ -> model
    SaveCustomShape -> case model.levelList of
        Editor (lvlInfo, ShapeCreator cshape) -> let newlvlInfo = {lvlInfo | customShapes = makeCShapeIndex lvlInfo.customShapes cshape}
            in {model | levelList = Editor (newlvlInfo, ShapeCreator ({colour = "#000000", canRotate = False, polygonPoints = [], coordinates = [], shortHand = ""}, Nothing))}
        _ -> model
    NewPiece -> case model.levelList of
        Editor (lvlInfo, ShapeCreator cshape) -> {model | levelList = Editor (lvlInfo, ShapeCreator ({colour = "#000000", canRotate = False, polygonPoints = [], coordinates = [], shortHand = ""}, Nothing))}
        _ -> model
    LoadShape i -> case model.levelList of
        Editor (lvlInfo, ShapeCreator cshape) -> case List.Extra.getAt (i-1) lvlInfo.customShapes of
            Just (shape, index) -> {model | levelList = Editor (lvlInfo, ShapeCreator (shape, Just index))}
            Nothing -> model
        _ -> model
    RemoveShape i -> case model.levelList of
        Editor (lvlInfo, ShapeCreator cshape) ->
            let
                cShapes = if List.length lvlInfo.customShapes > 1 then List.Extra.removeAt (i-1) lvlInfo.customShapes else []
            in {model | levelList = Editor ({lvlInfo | customShapes = applyNumbers cShapes}, ShapeCreator ({colour = "#000000", canRotate = False, polygonPoints = [], coordinates = [], shortHand = ""}, Nothing))}
        Editor (lvlInfo, LevelNumber) ->
            let
                lvl = case List.Extra.getAt (model.currentLevel-1) lvlInfo.levels of
                    Just l -> l
                    Nothing -> {boardx = 0, boardy = 0, solution = []}
                newSol = {lvl | solution = List.Extra.removeAt (i-1) lvl.solution}
            in {model | levelList = Editor ({lvlInfo | levels = List.Extra.setAt (model.currentLevel-1) newSol lvlInfo.levels}, LevelNumber)}
        _ -> model
    RemoveTetroShape shape -> case model.levelList of
        Editor (lvlInfo, LevelNumber) ->
            let
                lvl = case List.Extra.getAt (model.currentLevel-1) lvlInfo.levels of
                    Just l -> l
                    Nothing -> {boardx = 0, boardy = 0, solution = []}
                newSol = {lvl | solution = List.Extra.remove (getShortHandA shape) lvl.solution}
            in {model | levelList = Editor ({lvlInfo | levels = List.Extra.setAt (model.currentLevel-1) newSol lvlInfo.levels}, LevelNumber)}
        _ -> model
    ToEditor -> case model.levelList of
        Editor (lvlInfo, LevelNumber) -> {model | levelList = Editor (lvlInfo, ShapeCreator ({colour = "#000000", canRotate = False, polygonPoints = [], coordinates = [], shortHand = ""}, Nothing)), currentLevel = 0}
        _ -> model
    ChangeSizeX s -> case model.levelList of
        Editor (lvlInfo, LevelNumber) -> let gmlevel = List.Extra.getAt (model.currentLevel-1) lvlInfo.levels in case gmlevel of
            Just lvl -> let newSize = String.toInt s in case newSize of
                Nothing -> model
                Just size -> {model | levelList = Editor ({lvlInfo | levels = List.Extra.setAt (model.currentLevel-1) {lvl | boardx = size} lvlInfo.levels}, LevelNumber)}
            Nothing -> model
        _ -> model
    ChangeSizeY s -> case model.levelList of
        Editor (lvlInfo, LevelNumber) -> let gmlevel = List.Extra.getAt (model.currentLevel-1) lvlInfo.levels in case gmlevel of
            Just lvl -> let newSize = String.toInt s in case newSize of
                Nothing -> model
                Just size -> {model | levelList = Editor ({lvlInfo | levels = List.Extra.setAt (model.currentLevel-1) {lvl | boardy = size} lvlInfo.levels}, LevelNumber)}
            Nothing -> model
        _ -> model
    AddShape i -> case model.levelList of
        Editor (lvlInfo, LevelNumber) -> case List.Extra.getAt (i-1) lvlInfo.customShapes of
            Just (shape, index) -> let mlvl = List.Extra.getAt (model.currentLevel-1) lvlInfo.levels in case mlvl of
                Nothing -> model
                Just lvl -> {model | levelList = Editor ({lvlInfo | levels = List.Extra.setAt (model.currentLevel-1) {lvl | solution = lvl.solution++[shape.shortHand]} lvlInfo.levels}, LevelNumber)}
            Nothing -> model
        _ -> model
    AddTetroShape shape -> case model.levelList of
        Editor (lvlInfo, LevelNumber) -> let mlvl = List.Extra.getAt (model.currentLevel-1) lvlInfo.levels in case mlvl of
                Just lvl -> {model | levelList = Editor ({lvlInfo | levels = List.Extra.setAt (model.currentLevel-1) {lvl | solution = lvl.solution++[getShortHandA shape]} lvlInfo.levels}, LevelNumber)}
                Nothing -> model
        _ -> model
    ChangeHeight h -> {model | windowHeight = Just h}
    CustomLevelsLoaded content -> { model | customLevels = content }
    _ -> model
