module JsonFunctions exposing (rootEncoder, rootDecoder, mousePosition)

import Json.Decode exposing (list)
import Json.Encode as E
import Types exposing (..)
import List exposing (map)

rootEncoder : EditLevelInfo -> E.Value
rootEncoder lvlInfo = E.object [
        ("customshapes", E.list encodeCShape (map Tuple.first lvlInfo.customShapes)),
        ("levels", E.list encodeLevel lvlInfo.levels)
    ]

encodeLevel : EditLevel -> E.Value
encodeLevel lvl = E.object [
        ("boardx", E.int lvl.boardx),
        ("boardy", E.int lvl.boardy),
        ("solution", E.list E.string (lvl.solution))
    ]

encodeCShape : CustomShape -> E.Value
encodeCShape shape = E.object [
        ("canrotate", E.string (boolToString shape.canRotate)),
        ("shorthand", E.string shape.shortHand),
        ("colour", E.string shape.colour),
        ("coordinates", E.list encodeTuple shape.coordinates),
        ("polygonpoints", E.list encodeTuple shape.polygonPoints)
    ]

encodeTuple : (Int, Int) -> E.Value
encodeTuple (x,y) = E.object [
        ("x", E.int x),
        ("y", E.int y)
    ]

rootDecoder : Json.Decode.Decoder LevelInfo
rootDecoder = Json.Decode.map2 LevelInfo
    (Json.Decode.maybe (Json.Decode.field "customshapes" <| Json.Decode.list customShapeDecoder))
    (Json.Decode.field "levels" <| Json.Decode.list levelDecoder)


customShapeDecoder : Json.Decode.Decoder CustomShape
customShapeDecoder = Json.Decode.map5 CustomShape
    (Json.Decode.field "canrotate" <| stringBoolDecoder)
    (Json.Decode.field "shorthand" <| Json.Decode.string)
    (Json.Decode.field "coordinates" <| Json.Decode.list (tupleDecoder))
    (Json.Decode.field "polygonpoints" <| Json.Decode.list (tupleDecoder))
    (Json.Decode.field "colour" <| Json.Decode.string)

tupleDecoder : Json.Decode.Decoder (Int, Int)
tupleDecoder = Json.Decode.map2 (\x y -> (x,y))
    (Json.Decode.field "x" <| Json.Decode.int)
    (Json.Decode.field "y" <| Json.Decode.int)

stringBoolDecoder : Json.Decode.Decoder Bool
stringBoolDecoder =
  (Json.Decode.andThen) (\val ->
    case val of
      "true" -> Json.Decode.succeed True
      "false" -> Json.Decode.succeed False
      _ -> Json.Decode.fail <| "Expecting \"true\" or \"false\" but found " ++ val) Json.Decode.string


levelDecoder : Json.Decode.Decoder Level
levelDecoder = 
    Json.Decode.map3 Level
        (Json.Decode.maybe (Json.Decode.field "boardx" <| Json.Decode.int))
        (Json.Decode.maybe (Json.Decode.field "boardy" <| Json.Decode.int))
        (Json.Decode.field "solution" <| Json.Decode.list Json.Decode.string)

boolToString : Bool -> String
boolToString b = case b of
    True -> "true"
    False -> "false"

mousePosition : Json.Decode.Decoder MousePosition
mousePosition = Json.Decode.map2 MousePosition
        (Json.Decode.at [ "detail", "x" ] Json.Decode.float)
        (Json.Decode.at [ "detail", "y" ] Json.Decode.float)
