module Main exposing (main)

import Browser exposing (UrlRequest)
import Browser.Navigation as Nav
import Html exposing (Html, section, div, a, text)
import Html.Attributes exposing (attribute, class, href)
import Html.Attributes
import Url exposing (Url)
import List exposing (map)
import List.Extra
import Json.Decode exposing (list)
import Json.Encode as E
import Http
import Grid exposing (Grid)
import File exposing (File)
import File.Select as Select
import File.Download as Download
import Task
import Browser.Events
import Browser.Dom
import Text exposing (exampleJsonText)
import Types exposing (Model, Msg(..), LevelList(..), EditState(..))
import JsonFunctions exposing (rootEncoder, rootDecoder)
import TypeHelperFunctions exposing (getHeight, createEmptyShape, toAbstractShapes, getRealShapes, levelToGrid, getLength)
import GameLogic exposing (setCurrentLevel, update)
import UI exposing (selectSectionColour, printNavigation, createBody, createTitle, createModal)

main =
    Browser.application
        { init = init
        , view = viewDocument
        , update = updateCmd
        , subscriptions = subscriptions
        , onUrlRequest = onUrlRequest
        , onUrlChange = onUrlChange
        }

onUrlRequest : UrlRequest -> Msg
onUrlRequest x = LinkClicked x

onUrlChange : Url -> Msg
onUrlChange x = UrlChanged x

viewDocument : Model -> Browser.Document Msg
viewDocument model = Browser.Document "Tetromino-Puzzle" [(view model)]

init : () -> Url -> Nav.Key -> ( Model, Cmd Msg )
init _ u k = let model = {windowHeight = Nothing, drawState = Nothing, exitModalState = False, darkMode = False, exampleResult = Nothing, exampleLevels = exampleJsonText, jsonHelpState = Nothing, customLevels = "", completedLevels = [], currentLevel = 0, levelList = Empty, mousePiecePosition = (0,0), mpos = (0,0), url=u, key=k, shapeStatus = createEmptyShape, saved = [],boardGrid = Grid.repeat 8 8 False, boardIsFull = False} in (model,
    Task.perform (ChangeHeight << round << .height << .viewport) Browser.Dom.getViewport)

subscriptions : Model -> Sub Msg
subscriptions model = Browser.Events.onResize (\w h -> ChangeHeight h)

view : Model -> Html Msg
view model = section ((selectSectionColour model)++[attribute "style" ("height: "++String.fromInt (getHeight model)++"px; overflow: hidden;")]) [
        div [] [createTitle model],
        printNavigation model,
        div [class "hero-head"] (createBody model),
        createModal model,
        div [class "hero-body max-height"] [],
        div [class "hero-foot"] [a [Html.Attributes.target "_blank", class "custom-link", href "https://gitlab.informatik.uni-halle.de/karl.schmidt2/elm-tetromino"] [text "Quellcode"]]
    ]

updateCmd : Msg -> Model -> (Model, Cmd Msg)
updateCmd msg model = case msg of
    LinkClicked urlRequest -> case urlRequest of
        Browser.Internal url -> ( model, Nav.pushUrl model.key (Url.toString url) )
        Browser.External href -> ( model, Cmd.none)
    UrlChanged url -> handleFragments model url
    LoadLevels -> (model, Http.get {
          url = "./solutions.json"
        , expect = Http.expectJson GotLevels rootDecoder
        })
    GotLevels l -> case l of
        Ok levels -> let
                        model1 = {model | levelList = Levels  (map ((toAbstractShapes levels.customShapes) >> getRealShapes >> (levelToGrid)) levels.levels)} in (setCurrentLevel model1 0, Nav.pushUrl model.key "/spiel#1")
        e -> ({model | levelList = Failed}, Cmd.none)
    LoadCustomLevels -> ( model, Select.file ["application/json"] CustomLevelsSelected)
    CustomLevelsSelected file -> ( model, Task.perform CustomLevelsLoaded (File.toString file))
    LoadExample -> let decoded = Json.Decode.decodeString rootDecoder model.exampleLevels in case decoded of
        Ok levels -> let
                        model1 = {model | levelList = Levels  (map ((toAbstractShapes levels.customShapes) >> getRealShapes >> (levelToGrid)) levels.levels)}
                    in (setCurrentLevel model1 0, Nav.pushUrl model.key "/spiel#1")
        e -> ({model | exampleResult = Just False}, Cmd.none)
    DecodeCustomLevels -> let decoded = Json.Decode.decodeString rootDecoder model.customLevels in case decoded of
        Ok levels -> let
                        model1 = {model | levelList = Levels  (map ((toAbstractShapes levels.customShapes) >> getRealShapes >> (levelToGrid)) levels.levels)} in (setCurrentLevel model1 0, Nav.pushUrl model.key "/spiel#1")
        e -> ({model | levelList = Failed}, Cmd.none)
    SaveLevels -> case model.levelList of
        Editor (lvlInfo, _) -> (model, Download.string "level.json" "application/json" (E.encode 4 (rootEncoder lvlInfo)))
        _ -> (model, Cmd.none)
    DeleteLevel -> case model.levelList of
        Editor (lvlInfo, LevelNumber) -> let mlvl = List.Extra.getAt (model.currentLevel-1) lvlInfo.levels in case mlvl of
                Nothing -> (model, Cmd.none)
                Just lvl -> ({model | levelList = Editor ({lvlInfo | levels = List.Extra.remove lvl lvlInfo.levels},
                    if List.length lvlInfo.levels == 1 || model.currentLevel == 1 then ShapeCreator ({canRotate = False, colour="#000000", polygonPoints = [], coordinates = [], shortHand = ""}, Nothing) else LevelNumber), currentLevel = model.currentLevel-1}, Nav.replaceUrl model.key ("/editor#"++String.fromInt (model.currentLevel - 1)))
        _ -> (model, Cmd.none)
    EnterEditor -> ({model | levelList = Editor ({customShapes = [], levels = []}, ShapeCreator ({canRotate = False, colour = "#000000", shortHand = "", polygonPoints = [], coordinates = []}, Nothing))}, Nav.pushUrl model.key "/editor#0")
    _ -> (update msg model, Cmd.none)

handleFragments : Model -> Url -> (Model, Cmd Msg)
handleFragments model u = case u.path of
    "/editor" -> case u.fragment of
        Just num -> case String.toInt num of
            Nothing -> (model, Cmd.none)
            Just intNum -> case model.levelList of
                Editor (lvlInfo, estate) -> if (intNum > List.length lvlInfo.levels) then
                    let
                        newInfo = {lvlInfo | levels = lvlInfo.levels++[{boardx=8, boardy=8, solution = []}]}
                    in ({model | levelList = Editor (newInfo, LevelNumber), currentLevel = List.length newInfo.levels}, Cmd.none)
                    else if (intNum > 0) then ({model | levelList = Editor (lvlInfo, LevelNumber), currentLevel = intNum}, Cmd.none)
                    else case estate of
                        ShapeCreator x -> ({model | levelList = Editor (lvlInfo, estate)}, Cmd.none)
                        LevelNumber -> ({model | levelList = Editor (lvlInfo, ShapeCreator ({canRotate = False, colour = "#000000", shortHand = "", coordinates = [], polygonPoints = []}, Nothing))}, Cmd.none)
                _ -> (update EnterEditor model, Cmd.none)
        _ -> (update EnterEditor model, Cmd.none)
    "/spiel" -> case u.fragment of
        Just num -> case String.toInt num of
            Nothing -> (model, Cmd.none)
            Just intNum -> case model.levelList of
                Levels l -> if intNum > 0 && intNum <= getLength model.levelList then (setCurrentLevel model (intNum-1), Cmd.none) else (model, Cmd.none)
                _ -> (update LoadLevels model, Cmd.none)
        _ -> (update ToMainMenu model, Cmd.none)
    _ -> (update ToMainMenu model, Cmd.none)
