module UI exposing (..)

import Html exposing (Html, Attribute, h1, div, input, text, button, a, span, p, section, footer, header)
import Html.Attributes exposing (disabled, href, shape, attribute, class, type_)
import Html.Events exposing (onClick, onInput)
import Svg exposing (Svg, svg, rect, polygon, pattern)
import Svg.Attributes exposing (viewBox, width, height)
import Svg.Events
import Json.Decode
import List exposing (map)
import Grid
import List.Extra
import Types exposing (..)
import GameLogic exposing (..)
import Text exposing (..)
import TypeHelperFunctions exposing (..)
import GeneralHelperFunctions exposing (..)
import JsonFunctions exposing (mousePosition)
import Tetros exposing (tetroShapes)
import Html.Attributes exposing (checked)

createTitle : Model -> Html Msg
createTitle model = case model.levelList of
    Levels l -> p [class "title is-1"] [span [onClick EnableModal, class "pointer"] [text "Tetromino-Puzzle"],input [class "mr-4 sswitch", attribute "style" "float: right;", attribute "type" "checkbox", onClick ToggleTheme, Html.Attributes.checked model.darkMode] []]
    Editor (linfo, estate) -> case estate of
        ShapeCreator shape -> p [class "title is-1"] [button [attribute "style" "float: left;", class "button is-success ml-4 mt-4", onClick SaveLevels] [text "💾"], span [onClick EnableModal, class "pointer"] [text "Formen erstellen"],input [class "mr-4 sswitch", attribute "style" "float: right;", attribute "type" "checkbox", onClick ToggleTheme, Html.Attributes.checked model.darkMode] []]
        LevelNumber -> p [class "title is-1"] [button [attribute "style" "float: left;", class "button is-success ml-4 mt-4", onClick SaveLevels] [text "💾"], span [onClick EnableModal, class "pointer"] [text "Level erstellen"],input [class "mr-4 sswitch", attribute "style" "float: right;", attribute "type" "checkbox", onClick ToggleTheme, Html.Attributes.checked model.darkMode] []]
    _ -> p [class "title is-1"] [text "Tetromino-Puzzle",input [class "mr-4 sswitch", attribute "style" "float: right;", attribute "type" "checkbox", onClick ToggleTheme, Html.Attributes.checked model.darkMode] []]

selectSectionColour : Model -> List (Attribute (Msg))
selectSectionColour model = if not (levelIncomplete model 0) then
    [class "hero has-text-centered is-fullheight is-success"] else if model.darkMode then [class "hero has-text-centered is-fullheight", attribute "data-theme" "dark"]
        else [class "hero has-text-centered is-fullheight is-white", attribute "data-theme" "light"]

createBody : Model -> List (Html Msg)
createBody model = case model.levelList of
    Levels _ -> [div [attribute "style" "display: flex; height: 100vh;", class "mr-6 ml-6"] [
        div [attribute "style" "display: flex"]
            (printCompleteButton model),
            svg (svgAttributes model) (boardPattern (Grid.width model.boardGrid) (Grid.height model.boardGrid)
            ++(printBoardSpace model)++(printBoard (Grid.width model.boardGrid) (Grid.height model.boardGrid))++(List.concat (map (printShape True) model.saved))
            ++(printShape False model.shapeStatus))]]
    Editor x -> printEditor model x
    _ -> createStartMenu model

createModal : Model -> Html Msg
createModal model =
    div [class (if model.exitModalState then "modal is-active fade" else "modal is-active fade fout")] [
        div [class "modal-background", onClick DisableModal] [],
        div [class "modal-card"] [
            header [class "modal-card-head"] [
                p [class "modal-card-title"] [text "Aktueller Levelfortschritt wird zurückgesetzt!"]
            ],
            section [class "modal-card-body"] [
                div [] [
                    h1 [] [text "Zurück zum Hauptmenü?"]
                ]
            ],
            footer [class "modal-card-foot"] [
                div [attribute "style" "float: left;"] [
                    a [href "/"] [button [class "button is-danger mr-4"] [text "Ja"]],
                    button [class "button is-warning", onClick DisableModal] [text "Abbrechen"]
                ]
            ]
        ]
    ]

-- Main Menu

createStartMenu : Model -> List (Html Msg)
createStartMenu model = case model.levelList of
    Empty -> [div [class "mt-6"] [button [class "button is-link", onClick LoadLevels] [text "Spiel Starten"]]]++customLevels model
    Loading -> [div [class "mt-6"] [button [class "button is-link is-loading"] [text "Spiel Starten"]]]++customLevels model
    Failed -> [div [class "mt-6"] [button [class "button is-danger", onClick LoadLevels] [text "Laden Fehlgeschlagen"]]]++customLevels model
    _ -> [span [] []]

customLevels : Model -> List (Html Msg)
customLevels model = [div [class "mt-4 mb-4"] [button [class "mr-2 button", onClick LoadCustomLevels] [text "Eigene Level hochladen"],
    button [class "ml-2 button is-link", onClick DecodeCustomLevels] [text "Eigene Level starten"]]]
    ++[div [] [button [class "button is-warning mr-2", onClick SwitchJsonHelp] [text "Hilfe"],
    button [class "button is-link ml-2", onClick EnterEditor] [text "Level-Editor"]]]++printJsonHelp model

printJsonHelp : Model -> List (Html Msg)
printJsonHelp model = case model.jsonHelpState of
    Just True -> [div [class "mt-4 content animate__animated animate__fadeIn"] (jsonHelpContent model)]
    Just False -> [div [class "mt-4 content animate__animated animate__fadeOut", attribute "style" "pointer-events: none;"] (jsonHelpContent model)]
    Nothing -> []

jsonHelpContent : Model -> List (Html Msg)
jsonHelpContent model = [div [] [text jsonTutorial1],
    div [] [text jsonTutorial2],
    div [] [text jsonTutorial3],
    div [] [text jsonTutorial4],
    Html.article [class "message mr-6 ml-6"] [
        div [class "message-header"] [text "beispiel.json"],
        div [class "message-body"] [div [] [Html.textarea [class "textarea", onInput ExampleChanged] [text exampleJsonText]],
        div [] (printExampleButton model)]]
    ]

printExampleButton : Model -> List (Html Msg)
printExampleButton model = case model.exampleResult of
    Nothing -> [button [class "button is-success mt-4", onClick LoadExample] [text "Beispiel ausprobieren"]]
    Just False -> [button [class "button is-danger mt-4", onClick LoadExample] [text "Ungültiges Json!"]]
    _ -> []

-- Game Board

printNavigation : Model -> Html Msg
printNavigation model = case model.levelList of
    Levels l -> div [class "mr-6 ml-6 mb-4"] [
        div [] [printPreviousButton model],
        div [] [printNextButton model False],
        div [] [button [class "button is-link is-rounded"] [text ("Level "++(String.fromInt model.currentLevel)++"/"++(String.fromInt (getLength model.levelList)))]]
        ]
    Editor (linfo, estate) -> case estate of
        ShapeCreator _ -> div [class "mr-6 ml-6 mb-4"] [
            div [] [printPreviousButton model],
            div [] [printNextButton model True],
            div [] [button [class "button is-link is-rounded"] [text "Teile bearbeiten"]]
            ]
        LevelNumber -> div [class "mr-6 ml-6 mb-4"] [
            div [] [printPreviousButton model],
            div [] [printNextButton model True],
            div [] [button [class "button is-link is-rounded"] [text ("Level "++(String.fromInt model.currentLevel)++"/"++(String.fromInt (getLength model.levelList)))]]
            ]
    _ -> span [] []

printPreviousButton : Model -> Html Msg
printPreviousButton model = case model.levelList of
    Levels l -> case model.currentLevel of
        1 -> button [class "button is-info is-rounded is-left is-disabled", disabled True] [text "Vorheriges Level"]
        _ -> a [href ("#"++(String.fromInt (model.currentLevel-1)))] [button [class "button is-info is-rounded is-left"] [text "Vorheriges Level"]]
    Editor (linfo, estate) -> case estate of
        ShapeCreator _ -> button [class "button is-info is-rounded is-left is-disabled", disabled True] [text "Zum Editor"]
        LevelNumber -> case model.currentLevel of
            1 -> a [href ("#"++(String.fromInt (model.currentLevel-1)))] [button [class "button is-info is-rounded is-left"] [text "Zum Editor"]]
            _ -> a [href ("#"++(String.fromInt (model.currentLevel-1)))] [button [class "button is-info is-rounded is-left"] [text "Vorheriges Level"]]
    _ -> span [] []

printNextButton : Model -> Bool -> Html Msg
printNextButton model always = case model.levelList of
    Levels l -> if (not always) && (levelIncomplete model 0 || model.currentLevel == getLength model.levelList) then button [class "button is-info is-rounded is-right is-disabled", disabled True] [text "Nächstes Level"]
        else a [href ("#"++(String.fromInt (model.currentLevel+1)))] [button [class "button is-info is-rounded is-right"] [text "Nächstes Level"]]
    Editor (lvlInfo, estate) -> a [href ("#"++(String.fromInt (model.currentLevel+1)))] [button [class "button is-info is-rounded is-right"] [text (if model.currentLevel >= List.length lvlInfo.levels then "Neues Level" else "Nächstes Level")]]
    _ -> span [] []

printCompleteButton : Model -> List (Html Msg)
printCompleteButton model = if levelIncomplete model 0 && model.boardIsFull then [div [attribute "class" "centerbutton"] [button [class "button is-success mt-4", onClick CompleteLevel] [text "Level abschließen!"]]]
    else []

printBoardSpace : Model -> List (Svg Msg)
printBoardSpace model = [rect [width "100%", height "100%", attribute "fill" (if model.darkMode then "#2a3439" else "#e0e0e0")] [text "Hallo"]]

boardPattern : Int -> Int -> List (Svg Msg)
boardPattern x y = [Svg.defs [] [pattern [attribute "id" "brect", viewBox "0 0 100 100", width (String.fromFloat (100/(toFloat x))++"%"),  height (String.fromFloat (100/(toFloat y))++"%")] [
    rect [attribute "class" "boardrect", attribute "x" "0", attribute "y" "0", attribute "width" "100", attribute "height" "100"] []
    ]]]

printBoard : Int -> Int -> List (Svg Msg)
printBoard x y = let (bx, by) = getBoardBeginCoordinates (x,y) in
    [rect [attribute "x" (String.fromInt bx), attribute "y" (String.fromInt by), attribute "width" (String.fromInt (x*50)), attribute "height" (String.fromInt (y*50)), attribute "fill" "url(#brect)"] []]

svgAttributes : Model -> List (Attribute Msg)
svgAttributes model =
    let
        contextMenuAction = case model.drawState of
            Just DPolygon -> UndoPolygonPoint
            _ -> RightRotate
    in [attribute "style" "height:80%; width:100%", viewBox "0 0 1000 500", onMouseMove, onContextMenu contextMenuAction]++(isMoving model)

makeTake : Int -> Bool -> List (Attribute Msg)
makeTake i take = if take then [onClick (TakeShape i)] else [onClick PlaceShape, attribute "cursor" "none"]

svgTransform : (Int, Int) -> Int -> Attribute Msg
svgTransform (x,y) v = attribute "transform" ("rotate("++(String.fromInt (v*90))++" "++(String.fromInt (x))++" "++(String.fromInt (y))++")")

printShape : Bool -> ShapeStatus -> List (Svg Msg)
printShape take s = case s.sType of
    S -> [polygon ((makeTake s.sIdent take)++[attribute "fill" "green", attribute "stroke" "#000000", (svgTransform (s.x,s.y) s.v),
        attribute "points" (String.concat (map (singlePoint s.x s.y) (getShapePoints S)))]) []]
    O -> [polygon ((makeTake s.sIdent take)++[attribute "fill" "yellow", attribute "stroke" "#000000",
        attribute "points" (String.concat (map (singlePoint s.x s.y) (getShapePoints O)))]) []]
    I -> [polygon ((makeTake s.sIdent take)++[attribute "fill" "cyan", attribute "stroke" "#000000", (svgTransform (s.x,s.y) s.v),
        attribute "points" (String.concat (map (singlePoint s.x s.y) (getShapePoints I)))]) []]
    Z -> [polygon ((makeTake s.sIdent take)++[attribute "fill" "red", attribute "stroke" "#000000", (svgTransform (s.x,s.y) s.v),
        attribute "points" (String.concat (map (singlePoint s.x s.y) (getShapePoints Z)))]) []]
    T -> [polygon ((makeTake s.sIdent take)++[attribute "fill" "purple", attribute "stroke" "#000000", (svgTransform (s.x,s.y) s.v),
        attribute "points" (String.concat (map (singlePoint s.x s.y) (getShapePoints T)))]) []]
    L -> [polygon ((makeTake s.sIdent take)++[attribute "fill" "orange", attribute "stroke" "#000000", (svgTransform (s.x,s.y) s.v), 
        attribute "points" (String.concat (map (singlePoint s.x s.y) (getShapePoints L)))]) []]
    J -> [polygon ((makeTake s.sIdent take)++[attribute "fill" "blue", attribute "stroke" "#000000", (svgTransform (s.x,s.y) s.v),
        attribute "points" (String.concat (map (singlePoint s.x s.y) (getShapePoints J)))]) []]
    Custom c -> [polygon ((makeTake s.sIdent take)++[attribute "fill" c.colour, attribute "stroke" "#000000"]++ (if c.canRotate then [svgTransform (s.x,s.y) s.v] else [])++[
        attribute "points" (String.concat (map (singlePoint s.x s.y) c.polygonPoints))]) []]
    _ -> []

-- Editor

printEditor : Model -> (EditLevelInfo, EditState) -> List (Html Msg)
printEditor model (levelInfo, editState) = case editState of
    LevelNumber -> let cl = List.Extra.getAt (model.currentLevel-1) levelInfo.levels in case cl of
        Nothing -> []
        Just el -> [div [attribute "style" "display: flex", class "mr-6 ml-6"] [
            div [class "mr-4"] [div [class "mb-4"] [button [class "button is-danger", onClick DeleteLevel] [text "Level löschen"]],
                div [class "mb-4"] [p [] [text "Breite: "], input [onInput ChangeSizeX, type_ "range", attribute "min" "1", attribute "max" "12", Html.Attributes.value (getWidthE model)] []],
                div [class "mb-4"] [p [] [text "Höhe: "], input [onInput ChangeSizeY, type_ "range", attribute "min" "1", attribute "max" "10", Html.Attributes.value (getHeightE model)] []],
                div [attribute "style" ("height:"++if model.windowHeight == Nothing then "400" else String.fromInt (round ((toFloat (getHeight model)) / (2.5)))++"px; overflow-x:hidden; overflow-y:auto;")] (map (drawLevelShape model.darkMode) ((map customToAbstract levelInfo.customShapes)++tetroShapes))
                ],
            svg (svgAttributes model) (boardPattern el.boardx el.boardy
            ++(printBoardSpace model)++(printBoard el.boardx el.boardy)++(printProgress el ((map customToAbstract levelInfo.customShapes)++tetroShapes))++
            [Svg.text_ [attribute "class" "unselectable", attribute "x" "20", attribute "y" "20"] [text ((String.fromInt el.boardx)++" x "++(String.fromInt el.boardy))]]),
            div [class "ml-4", attribute "style" ("height:"++if model.windowHeight == Nothing then "700" else String.fromInt (round ((toFloat (getHeight model)) / (1.5)))++"px; overflow-x:hidden; overflow-y:auto;")] (if List.length (getLevelSolution model.currentLevel levelInfo.levels) > 0 then (map (drawSavedShape model.darkMode) ((map (getShapeFromShortHand (map (Tuple.first << customToAbstract) levelInfo.customShapes)) (List.Extra.zip (List.range 1 (List.length (getLevelSolution model.currentLevel levelInfo.levels))) (getLevelSolution model.currentLevel levelInfo.levels))))) else [drawSkeletonShape model.darkMode])]]
    ShapeCreator shape -> [div [attribute "style" "display: flex", class "mr-6 ml-6"] [
        div [class "mr-4"] [div [class "mb-4"] [button [class "button is-info", onClick NewPiece] [text "neues Puzzleteil"]], 
                div [class "mb-4"] [Html.label [class "checkbox", Html.Events.onCheck CheckRotate] [input [type_ "checkbox", checked (getRotatable model)] [], text "Rotierbar"]],
                div [class "mb-4"] [text "Farbe", input [type_ "color", onInput NewColour, Html.Attributes.value (getColour model)] []],
                div [class "mb-4"] [text "Name", input [onInput ChangeName, Html.Attributes.value (getShapeName model)] []],
                div [class "mb-4"] [button [class "button is-success", onClick DrawCoordinates] [text "Koordinaten zeichnen"]],
                div [class "mb-4"] [button [class "button is-success", onClick DrawPolygon] [text "Polygon zeichnen"]],
                div [class "mb-4"] [button [class "button is-danger", onClick ResetPolygon] [text "Polygon löschen"]],
                div [class "mb-4"] [button [class "button is-link", onClick SaveCustomShape] [text "Puzzleteil speichern"]]
                ],
            svg (svgAttributes model) (boardPattern 8 8
            ++(printBoardSpace model)++(printBoard 8 8)++(printDrawState model)),
            div [attribute "style" ("height:"++if model.windowHeight == Nothing then "700" else String.fromInt ((round ((toFloat (getHeight model)) / (1.5))))++"px; overflow-x:hidden; overflow-y:auto;"), class "ml-4"] (if List.length levelInfo.customShapes > 0 then (map (drawSavedShape model.darkMode) ((map customToAbstract levelInfo.customShapes))) else [drawSkeletonShape model.darkMode])]]

printProgress : EditLevel -> List (AbstractShape, Int) -> List (Svg Msg)
printProgress el shapes =
    let
        (bbx, bby) = getBoardBeginCoordinates (el.boardx, el.boardy)
        points = el.boardx * el.boardy
        occupiedPoints = List.sum (map (List.length << getAbstractCoordinates << Tuple.first << getShapeFromShortHand (map Tuple.first shapes)) (List.Extra.zip (List.range 1 (List.length el.solution)) el.solution))
    in [rect [attribute "x" (String.fromInt bbx), attribute "y" (String.fromInt bby), attribute "width" (String.fromInt (50*el.boardx)), attribute "height" (String.fromInt (round ((toFloat occupiedPoints/toFloat points)*toFloat (50*el.boardy)))),
        attribute "fill" (if occupiedPoints < points then "#ff0000" else if occupiedPoints==points then "#40ff00" else "#ffff00"),
        attribute "style" "opacity:0.5;"] [],
        Svg.text_ [attribute "class" "unselectable", attribute "x" "20", attribute "y" "480"] [text ((String.fromInt occupiedPoints)++" / "++(String.fromInt points))]
        ]

drawSkeletonShape : Bool -> Html Msg
drawSkeletonShape darkMode = div [] [
        svg [attribute "style" "height:90%; width:90%", viewBox "0 0 100 100"] [
            rect [attribute "class" (if darkMode then "shapeSelectDark" else "shapeSelectLight"), attribute "x" "0", attribute "y" "0", attribute "rx" "20", attribute "ry" "20", attribute "width" "100", attribute "height" "100", attribute "fill" (if darkMode then "#1a1a1a" else "#efefef")] []]]

drawSavedShape : Bool -> (AbstractShape, Int) -> Html Msg
drawSavedShape darkMode (ashape, index) = case ashape of
    Custom shape -> div [attribute "style" "cursor: pointer;"] [
        svg [attribute "style" "height:90%; width:90%", viewBox "0 0 100 100"] [
            rect [onClick (LoadShape index), attribute "class" (if darkMode then "shapeSelectDark" else "shapeSelectLight"), attribute "x" "0", attribute "y" "0", attribute "rx" "20", attribute "ry" "20", attribute "width" "100", attribute "height" "100", attribute "fill" (if darkMode then "#1a1a1a" else "#efefef")] [],
            polygon [attribute "class" "no-cursor" , attribute "stroke" "#000000", attribute "fill" shape.colour, attribute "points" (String.concat (map (singleScaledPoint (getBoundingBox shape.polygonPoints)) (pNormalize shape.polygonPoints)))] [],
            Svg.circle [onClick (RemoveShape index), attribute "class" "exit", attribute "cx" "92", attribute "cy" "8", attribute "r" "8", attribute "fill" "#ffa0a0"] [],
            Svg.text_ [attribute "class" "unselectable", attribute "style" "pointer-events: none;", attribute "fill" "#0f0f0f", attribute "x" "89", attribute "y" "12"] [text "x"]
            ]
        ]
    x -> div [attribute "style" "cursor: pointer;"] [
        svg [attribute "style" "height:90%; width:90%", viewBox "0 0 100 100"] [
            rect [attribute "class" (if darkMode then "shapeSelectDark" else "shapeSelectLight"), attribute "x" "0", attribute "y" "0", attribute "rx" "20", attribute "ry" "20", attribute "width" "100", attribute "height" "100", attribute "fill" (if darkMode then "#1a1a1a" else "#efefef")] [],
            polygon [attribute "class" "no-cursor" , attribute "stroke" "#000000", attribute "fill" (getShapeColour x), attribute "points" (String.concat (map (singleScaledPoint (getBoundingBox (getShapePoints x))) (pNormalize (getShapePoints x))))] [],
            Svg.circle [onClick (RemoveTetroShape x), attribute "class" "exit", attribute "cx" "92", attribute "cy" "8", attribute "r" "8", attribute "fill" "#ffa0a0"] [],
            Svg.text_ [attribute "class" "unselectable", attribute "style" "pointer-events: none;", attribute "fill" "#0f0f0f", attribute "x" "89", attribute "y" "12"] [text "x"]
            ]
        ]

drawLevelShape : Bool -> (AbstractShape, Int) -> Html Msg
drawLevelShape darkMode (ashape, index) = case ashape of
    Custom shape -> div [attribute "style" "cursor: pointer;"] [
        svg [attribute "style" "height:90%; width:90%", viewBox "0 0 100 100"] [
            rect [attribute "class" (if darkMode then "shapeSelectDark" else "shapeSelectLight"), onClick (AddShape index), attribute "x" "0", attribute "y" "0", attribute "rx" "20", attribute "ry" "20", attribute "width" "100", attribute "height" "100", attribute "fill" (if darkMode then "#1a1a1a" else "#efefef")] [],
                polygon [attribute "class" "no-cursor", attribute "stroke" "#000000", attribute "fill" shape.colour, attribute "points" (String.concat (map (singleScaledPoint (getBoundingBox shape.polygonPoints)) (pNormalize shape.polygonPoints)))] []
            ]
        ]
    x -> div [attribute "style" "cursor: pointer;"] [
        svg [attribute "style" "height:90%; width:90%", viewBox "0 0 100 100"] [
            rect [attribute "class" (if darkMode then "shapeSelectDark" else "shapeSelectLight"), onClick (AddTetroShape x), attribute "x" "0", attribute "y" "0", attribute "rx" "20", attribute "ry" "20", attribute "width" "100", attribute "height" "100", attribute "fill" (if darkMode then "#1a1a1a" else "#efefef")] [],
                polygon [attribute "class" "no-cursor", attribute "stroke" "#000000", attribute "fill" (getShapeColour x), attribute "points" (String.concat (map (singleScaledPoint (getBoundingBox (getShapePoints x))) (pNormalize (getShapePoints x))))] []
            ]
        ]

printDrawState : Model -> List (Svg Msg)
printDrawState model =
    let 
        gridCoordinates = [(0,0),(0,1),(0,2),(0,3),(0,4),(0,5),(0,6),(0,7),(1,0),(1,1),(1,2),(1,3),(1,4),(1,5),(1,6),(1,7),(2,0),(2,1),(2,2),(2,3),(2,4),(2,5),(2,6),(2,7),(3,0),(3,1),(3,2),(3,3),(3,4),(3,5),(3,6),(3,7),(4,0),(4,1),(4,2),(4,3),(4,4),(4,5),(4,6),(4,7),(5,0),(5,1),(5,2),(5,3),(5,4),(5,5),(5,6),(5,7),(6,0),(6,1),(6,2),(6,3),(6,4),(6,5),(6,6),(6,7),(7,0),(7,1),(7,2),(7,3),(7,4),(7,5),(7,6),(7,7)]
        (bbx, bby) = getBoardBeginCoordinates (8,8) 
    in case model.drawState of
            Nothing -> []
            Just DPolygon -> let gCoordinates = gridCoordinates++[(8,0),(8,1),(8,2),(8,3),(8,4),(8,5),(8,6),(8,7),(8,8),(0,8),(1,8),(2,8),(3,8),(4,8),(5,8),(6,8),(7,8),(8,8)] in
                drawCoordinates model++(map (drawPCircle (bbx, bby) model.mpos) gCoordinates)++(map (drawPCube (bbx, bby)) gCoordinates)++drawCurrentPolgyon model
            Just DCoordinates -> (map (drawCCube (bbx, bby) True) gridCoordinates)++drawCoordinates model

drawCurrentPolgyon : Model -> List (Svg Msg)
drawCurrentPolgyon model = let (bbx, bby) = getBoardBeginCoordinates (8,8) in case model.levelList of
    Editor (x, ShapeCreator (cshape,index)) -> [polygon [attribute "style" "pointer-events:none; opacity:0.6;", attribute "stroke" "#000000", attribute "fill" cshape.colour , attribute "points" (String.concat (map (singlePoint bbx bby) (cshape.polygonPoints))++
        let maybemouse = getBoardCoordinatesM 8 8 model.mpos in case maybemouse of
            Just (mx,my) -> (String.fromInt (bbx+mx*50))++","++(String.fromInt (bby+my*50))
            Nothing -> ""
        )] []]
    _ -> []

drawPCircle : (Int, Int) -> (Int, Int) -> (Int, Int) -> Svg Msg
drawPCircle (bx,by) (mx,my) (cx, cy) =
    let
        hx = abs (mx-(bx+(cx*50)))
        hy = abs (my-(by+(cy*50)))
        isHighlight = hx<25 && hy<25
    in if isHighlight then Svg.circle [onClick (AddPolygonPoint (cx,cy)), attribute "cx" (String.fromInt (cx*50+bx)), attribute "cy" (String.fromInt (cy*50+by)), attribute "r" "10",  attribute "style" "fill:yellow;stroke:black;stroke-width:1"] []
    else Svg.circle [onClick (AddPolygonPoint (cx,cy)), attribute "cx" (String.fromInt (cx*50+bx)), attribute "cy" (String.fromInt (cy*50+by)), attribute "r" "5",  attribute "style" "fill:yellow;stroke:black;stroke-width:1"] []

drawPCube : (Int, Int) -> (Int, Int) -> Svg Msg
drawPCube (bx,by) (cx, cy) = rect [onClick (AddPolygonPoint (cx,cy)), attribute "x" (String.fromInt (cx*50+bx-25)), attribute "y" (String.fromInt (cy*50+by-25)), attribute "width" "50", attribute "height" "50",  attribute "style" "opacity:0"] []

drawCCube : (Int, Int) -> Bool -> (Int, Int) -> Svg Msg
drawCCube (bx, by) addC (cx, cy) = if addC then rect [onClick (AddCoordinate (cx,cy)), attribute "x" (String.fromInt (cx*50+bx)), attribute "y" (String.fromInt (cy*50+by)), attribute "width" (String.fromInt 50), attribute "height" (String.fromInt 50), attribute "style" "fill:gray;stroke:black;stroke-width:3;fill-opacity:0.3;stroke-opacity:0.9"] []
    else rect [onClick (RemoveCoordinate (cx,cy)), attribute "x" (String.fromInt (cx*50+bx)), attribute "y" (String.fromInt (cy*50+by)), attribute "width" (String.fromInt 50), attribute "height" (String.fromInt 50), attribute "style" "fill:red;stroke:black;stroke-width:3;fill-opacity:1;stroke-opacity:1"] []

drawCoordinates : Model -> List (Svg Msg)
drawCoordinates model = let (bbx, bby) = getBoardBeginCoordinates (8,8) in case model.levelList of
    Editor (x, ShapeCreator (cshape,index)) -> map (drawCCube (bbx, bby) False) cshape.coordinates 
    _ -> []

-- Events

onMouseMove : Svg.Attribute Msg
onMouseMove  = Svg.Events.on "mousemoveWithCoordinates" (Json.Decode.map (\v -> Position (round v.x) (round v.y)) mousePosition)

onContextMenu : Msg -> Attribute Msg
onContextMenu message = Html.Events.on "auxclick" (Json.Decode.succeed message)

-- Helper Functions

isMoving : Model -> List (Attribute Msg)
isMoving model = case model.shapeStatus.sType of
    None -> []
    _ -> [onClick PlaceShape, attribute "cursor" "none"]

singlePoint : Int -> Int -> (Int, Int) -> String
singlePoint x y (px, py) = (String.fromInt (x+px*50))++","++(String.fromInt (y+py*50))++" "

singleScaledPoint : (Int, Int) -> (Int, Int) -> String
singleScaledPoint bb (px, py) =
    let
        bmax = maxTuple bb
        blockSize = 100//(bmax+2)
    in (String.fromInt ((50-((blockSize*Tuple.first bb)//2))+px*blockSize))++","++(String.fromInt ((50-((blockSize*Tuple.second bb)//2))+py*blockSize))++" "

getBoundingBox : List (Int,Int) -> (Int, Int)
getBoundingBox polygonPoints =
    let
        normalPoints = pNormalize polygonPoints
        (mx,my) = (List.maximum (map Tuple.first normalPoints),List.maximum (map Tuple.second normalPoints))
    in case mx of
        Just x -> case my of
            Just y -> (x,y)
            Nothing -> (0,0)
        Nothing -> (0,0)